require 'yaml'
require 'kwalify'

module Blufin

    class Config

        @@data = {}

        # Validate and initialize the configuration file.
        # @return void
        def self.init(schema_file, template_file, config_file, gem_name, setup_command = 'x')
            schema_file = File.expand_path(schema_file)
            template_file = File.expand_path(template_file)
            config_file = File.expand_path(config_file)
            raise RuntimeError, "File not found: #{schema_file}" unless Blufin::Files::file_exists(schema_file)
            raise RuntimeError, "File not found: #{template_file}" unless Blufin::Files::file_exists(template_file)
            if Blufin::Files::file_exists(config_file)
                # Validate the user created config file.
                load_config(config_file, schema_file, gem_name)
            else
                system('clear')
                if Blufin::Terminal::prompt_yes_no(" Missing Configuration File: #{Blufin::Terminal::format_directory(config_file)}", 'This script can automatically create it.', 'Create configuration file')
                    make_config(template_file, config_file)
                    edit_config(config_file)
                    Blufin::Terminal::custom('File Created', 56, "Now try running: #{Blufin::Terminal::format_command(gem_name)}", nil, false)
                end
                exit
            end
        end

        # Straight-up returns the data object so you can get keys using string literals.
        # Hard-coding keys is OK because configuration rarely changes and it's validated, so you can be assured the data is there.
        # @return Hash
        def self.get
            @@data
        end

        # Use for path. Basically wraps them in File.expand_path.
        # @return String
        def self.get_path(*args)
            val = get
            args.each { |arg| val = val[arg] }
            File.expand_path(val)
        end

        # Opens the config file in VIM.
        # @return void
        def self.edit_config(config_file)
            system("vim #{config_file}")
        end

        # This validates the configuration file against the defined schema.
        # @return Array
        def self.validate_file(config_file, schema_file)
            begin
                schema_file_parsed = YAML.load_file(schema_file)
            rescue => e
                Blufin::Terminal::error("Failed to parse schema file: #{Blufin::Terminal::format_directory(schema_file)}", e.message)
            end
            validator = Kwalify::Validator.new(schema_file_parsed)
            begin
                document = YAML.load_file(config_file)
                # This is here because the IDE is incorrectly resolving to Convoy::validate().
                # noinspection RubyArgCount
                errors = validator.validate(document)
            rescue => e
                Blufin::Terminal::error("Failed to parse config file: #{Blufin::Terminal::format_directory(config_file)}", e.message)
            end
            return document, errors
        end

        # This shows the schema validation errors in a consistent manner across all the gems.
        # @return void
        def self.display_errors(errors, gem_name, setup_command = 'x')
            errors_output = []
            errors.each { |e|
                errors_output << "[#{e.path}] #{e.message}"
            }
            invalid_configuration(gem_name, errors, setup_command)
        end

        # Standardized way of outputting invalid configuration.
        # @return void
        def self.invalid_configuration(gem_name, errors = nil, setup_command = 'x')
            Blufin::Terminal::error("Your configuration file is invalid. To fix it run: #{Blufin::Terminal::format_command("#{gem_name} #{setup_command}")}", errors, true)
        end

        private

        # Loads the config file into global Hash.
        # @return void
        def self.load_config(config_file, schema_file, gem_name, setup_command = 'x')
            document, errors = validate_file(config_file, schema_file)
            if errors && !errors.empty?
                display_errors(errors, gem_name, setup_command)
            else
                @@data = document
            end
        end

        # Creates a the config file from the given template.
        # @return
        def self.make_config(template_file, config_file)
            `cp #{template_file} #{config_file}`
        end

    end

end